# -*- coding: utf-8 -*-
"""
Created on Tue Mar 24 21:28:39 2026

@author: trent
"""

from threading import Thread

import typer
import matplotlib.pyplot as plt

from glyphloom.data.fifth_edition import SpellData_5e
# from glyphloom.generation.geometry import Founts, Leylines, SAFE_MATH
# from glyphloom.generation.draw import Glyph


def wait_for_input(flag):
    input("Press Enter to continue...")
    flag['done'] = True


app = typer.Typer(no_args_is_help=True)


@app.command()
def draw(spell: str = typer.Argument(None),
         spell_opt: str = typer.Option("Modify Memory", "-s", "--spell"),
         source_opt: str = typer.Option("offline", "-o", "--origin"),
         ritual = typer.Option(None, "--ritual",
                               help="Ritual override for display, options are 0, 1, False, True, false, true"),
         fount_expression: str = typer.Option("polygon", "-f", "--founts",
                                              help="'founts' must be a semicolon (';') separated list of expressions with 'domain' as the input values of the function(s)"),
         fount_min=typer.Option("0", "--fount-min",
                                help="'fount-min' must be a number representing the minimum value of the 'founts' parametric equation"),
         fount_max=typer.Option("1", "--fount-max",
                                help="'fount-max' must be a number representing the minimum value of the 'founts' parametric equation"),
         leyline_expression: str = typer.Option("linear",
                                                "-p", "--path",
                                                help="'path' must be a semicolon (';') separated list of expressions with either 'domain' or 'P,Q' as the input values of the function(s)"),
         leyline_min=typer.Option("0", "--leyline-min",
                                  help="'leyline-min' must be a number representing the minimum value of the 'path' parametric equation, only affects functions with 'domain' as the input values."),
         leyline_max=typer.Option("1", "--leyline-max",
                                  help="'leyline-max' must be a number representing the minimum value of the 'path' parametric equation, only affects functions with 'domain' as the input values."),
         leyline_resolution=typer.Option("150", "--path-resolution",
                                         help="'path-resolution' must be an integer representing the number of samples to take along the leyline for shaping."),
         show_legend: bool = typer.Option(False, "-l", "--show-legend",
                                          help="Boolean to show legend"),
         save_file: str = typer.Option(None, "--save-file",
                                       help="Relative or Absolute filepath for png output.")
         ) -> None:
    
    from glyphloom.generation.geometry import Founts, Leylines, SAFE_MATH
    from glyphloom.generation.draw import Glyph

    done = {"done": False}
    t = Thread(target=wait_for_input, args=(done,), daemon=True)
    t.start()

    spell = spell or spell_opt

    plt.ion()
    data = SpellData_5e.get_spell(spell, source_opt)
    if ritual is not None:
        if isinstance(ritual, bool):
            ritual = ritual
        elif ritual in ["0", "1"]:
            ritual = bool(int(ritual))
        elif ritual.lower().capitalize() in ['True', 'False']:
            ritual = eval(ritual.lower().capitalize())
        data.ritual = ritual

    # Unpack founts
    shape = tuple(fount_expression.split(";"))
    if len(shape) == 1:
        shape = shape[0]

    fount_min = eval(fount_min, {}, SAFE_MATH)
    fount_max = eval(fount_max, {}, SAFE_MATH)
    founts = Founts(n_points=13,
                    expression=shape,
                    domain_min=(fount_min),
                    domain_max=(fount_max))

    # Unpack Leylines
    path = tuple(leyline_expression.split(";"))
    if len(path) == 1:
        path = path[0]
    leyline_min = eval(leyline_min, {}, SAFE_MATH)
    leyline_max = eval(leyline_max, {}, SAFE_MATH)
    leyline_resolution = int(eval(leyline_resolution, {}, SAFE_MATH))

    leylines = Leylines(founts=founts,
                        expression=path,
                        domain_min=leyline_min,
                        domain_max=leyline_max,
                        resolution=leyline_resolution)
    fig = Glyph(data,
                leylines).draw(legend=show_legend)
    if save_file is not None:
        plt.savefig(save_file, transparent=True, dpi='figure')
    plt.show()

    while not done['done'] and plt.fignum_exists(fig.number):
        plt.pause(0.1)


@app.command()
def custom(spell_name:str = typer.Argument(""),
           spell_name_opt:str = typer.Option("", "-n", "--spell-name",
                                             help="Spell Name for custom input"),
           level:int = typer.Argument(0),
           level_opt:int = typer.Option(0, "-v", "--level",
                                        help="Integer level of the spell from 0-20"),
           school:str = typer.Argument(None),
           school_opt:str = typer.Option(None, "-c", "--school",
                                       help=f"School of the spell: {tuple(SpellData_5e.school.options)}"),
           damage_type:str = typer.Argument(None),
           damage_type_opt:str = typer.Option(None, "-d", "--damage-type",
                                              help=f"Damage type of the spell: {tuple(SpellData_5e.damage_type.options)}"),
           area_of_effect:str = typer.Argument(None),
           area_of_effect_opt:str = typer.Option(None, "-aoe", "--area-of-effect",
                                                 help=f"Area of effect of the spell: {tuple(SpellData_5e.area_of_effect.options)}"),
           range:str = typer.Argument(None),
           range_opt:str = typer.Option(None, "-r", "--spell-range",
                                          help=f"Range of the spell as described by the SRD: {tuple(SpellData_5e.range.options)}"),
           duration:str = typer.Argument(None),
           duration_opt:str = typer.Option(None, "-t", "--spell-duration",
                                           help=f"Duration of the spell as described by the SRD: {tuple(SpellData_5e.duration.options)}"),
           ritual = typer.Option("0", "--ritual",
                                 help="Ritual override for display, options are 0, 1, False, True, false, true"),
           fount_expression: str = typer.Option("polygon", "-f", "--founts",
                                                help="'founts' must be a semicolon (';') separated list of expressions with 'domain' as the input values of the function(s)"),
           fount_min=typer.Option("0", "--fount-min",
                                  help="'fount-min' must be a number representing the minimum value of the 'founts' parametric equation"),
           fount_max=typer.Option("1", "--fount-max",
                                  help="'fount-max' must be a number representing the minimum value of the 'founts' parametric equation"),
           leyline_expression: str = typer.Option("linear",
                                                  "-p", "--path",
                                                  help="'path' must be a semicolon (';') separated list of expressions with either 'domain' or 'P,Q' as the input values of the function(s)"),
           leyline_min=typer.Option("0", "--leyline-min",
                                    help="'leyline-min' must be a number representing the minimum value of the 'path' parametric equation, only affects functions with 'domain' as the input values."),
           leyline_max=typer.Option("1", "--leyline-max",
                                    help="'leyline-max' must be a number representing the minimum value of the 'path' parametric equation, only affects functions with 'domain' as the input values."),
           leyline_resolution=typer.Option("150", "--path-resolution",
                                           help="'path-resolution' must be an integer representing the number of samples to take along the leyline for shaping."),
           show_legend: bool = typer.Option(False, "-l", "--show-legend",
                                            help="Boolean to show legend"),
           save_file: str = typer.Option(None, "--save-file",
                                         help="Relative or Absolute filepath for png output.")
           ) -> None:
    
    from glyphloom.generation.geometry import Founts, Leylines, SAFE_MATH
    from glyphloom.generation.draw import Glyph
    
    done = {"done": False}
    t = Thread(target=wait_for_input, args=(done,), daemon=True)
    t.start()
    plt.ion()
    
    # Create Data
    ritual = False
    if ritual is not None:
        if isinstance(ritual, bool):
            ritual = ritual
        elif ritual in ["0", "1"]:
            ritual = bool(int(ritual))
        elif ritual.lower().capitalize() in ['True', 'False']:
            ritual = eval(ritual.lower().capitalize())
    data = SpellData_5e(name = spell_name or spell_name_opt,
                        level = level or level_opt,
                        school = school or school_opt,
                        damage_type = damage_type or damage_type_opt,
                        area_of_effect = area_of_effect or area_of_effect_opt,
                        range = range or range_opt,
                        duration = duration or duration_opt,
                        ritual = ritual)
    

    # Unpack founts
    shape = tuple(fount_expression.split(";"))
    if len(shape) == 1:
        shape = shape[0]

    fount_min = eval(fount_min, {}, SAFE_MATH)
    fount_max = eval(fount_max, {}, SAFE_MATH)
    founts = Founts(n_points=13,
                    expression=shape,
                    domain_min=(fount_min),
                    domain_max=(fount_max))

    # Unpack Leylines
    path = tuple(leyline_expression.split(";"))
    if len(path) == 1:
        path = path[0]
    leyline_min = eval(leyline_min, {}, SAFE_MATH)
    leyline_max = eval(leyline_max, {}, SAFE_MATH)
    leyline_resolution = int(eval(leyline_resolution, {}, SAFE_MATH))

    leylines = Leylines(founts=founts,
                        expression=path,
                        domain_min=leyline_min,
                        domain_max=leyline_max,
                        resolution=leyline_resolution)
    fig = Glyph(data,
                    leylines).draw(legend=show_legend)
    if save_file is not None:
        plt.savefig(save_file, transparent=True, dpi='figure')
    plt.show()

    while not done['done'] and plt.fignum_exists(fig.number):
        plt.pause(0.1)



# -*- coding: utf-8 -*-
"""
Created on Sat Mar 14 15:33:40 2026

@author: trent
"""

from types import MappingProxyType
import typing
import re
import copy
import inspect

import numpy as np
import matplotlib.pyplot as plt

SAFE_MATH = MappingProxyType({
    k: getattr(np, k) for k in ["sin", "cos", "tan", "arcsin", "arccos", 
                                "arctan", "arctan2", "sinh", "cosh", "tanh", 
                                "exp", "log", "log10", "sqrt", "abs", "fabs",
                                "floor", "ceil", "pi", "e", "mean", "hypot",
                                'array', 'linalg', 'min', 'max', 'linspace',
                                'arange', 'where', 'random']
    })

class Founts:
    """
    Generic method for generating the n and k possible lines given some 
    generative input.
    
    """
    
    PREDEFINED = MappingProxyType({
        'polygon' : ('radius', 
                     'start_angle',
                     'cw'),
        'golden' : tuple()})
  
    #---- Numpy overrides
    def __iter__(self):
        return iter(self.nodes)
    
    def __array__(self, dtype=None, copy=False):
        founts = self.nodes
        if dtype:
            founts = founts.astype(dtype)
        if copy:
            founts = founts.copy()
        return founts
    
    @property
    def shape(self):
        return self.nodes.shape
    
    @property
    def T(self):
        new = copy.deepcopy(self)
        new.founts = new.founts.T
        return new
    
    def __getitem__(self, item):
        if isinstance(item, typing.Iterable):
            return self.nodes[*item]
        return self.nodes[item]
    
    def __init__(self, 
                 nodes: np.ndarray = None,
                 n_points: int = 13,
                 expression: str | tuple[str] = None,
                 domain_min: float = -1,
                 domain_max: float = 1,
                 **kwargs):
        
        
        self.nodes = nodes
        self.n_points = n_points
        self.expression = expression
        self.domain_min = domain_min
        self.domain_max = domain_max
        self.kwargs = kwargs
        
        if nodes is None and not expression:
            self.nodes = type(self).polygon(n_points=n_points).nodes
            return       
        elif nodes is not None:
            self.nodes = np.array([*nodes])
            return
        
        if isinstance(expression, str) and expression not in self.PREDEFINED:
            # Simple Function
            domain = np.linspace(domain_min, domain_max, n_points)
            env = {'__builtins__': None, 
                   **SAFE_MATH, 
                   'domain': domain}
            ys = eval(expression, env)
            self.nodes = np.array([domain, ys])
            return
        elif isinstance(expression, str) and expression in self.PREDEFINED:
            func = getattr(type(self), expression, lambda *args, **kwargs: f'No function {expression}')
            res = func(nodes=nodes, n_points=n_points, expression=expression,
                       domain_min=domain_min, domain_max=domain_max, **kwargs)
            if isinstance(res, type(self)):
                self.nodes = res.nodes
            else:
                raise NotImplementedError(f"No predefined shape function for '{expression}'")
                
        elif isinstance(expression, tuple) and len(expression) == 2:
            # Parametric function
            domain = np.linspace(domain_min, domain_max, n_points)
            env = {'__builtins__': {}, **SAFE_MATH, 'domain': domain}
            xs = eval(expression[0], env)
            ys = eval(expression[1], env)
            self.nodes = np.array([xs, ys])
            return
    
    def __repr__(self):
        arg = [f'{attr_name}={getattr(self, attr_name)}' 
               for attr_name in list(getattr(self, '__dict__', {})) 
               if not re.search('__.*__', attr_name)]
        arg_str = ', '.join(arg)
        return f'{type(self).__name__}({arg_str})'
    
    @classmethod
    def polygon(cls, 
                n_points: int = 13,
                radius: float = 1,
                start_angle:float = None,
                cw: bool = True,
                **kwargs) -> typing.Self:
        """
        Default polygon generation, equivalent to 
        
        expression = ('radius*cos(start_angle + 2*pi/n_points)',
                      'radius*sin(start_angle + 2*pi/n_points)')
        where radius would be a float
              start_angle is a float
              n_points is an integer

        Parameters
        ----------
        n_points : int, optional
            Number of vertices. The default is 13.
        radius : float, optional
            Radius of the inscribed circle. The default is 1.
        start_angle : float, optional
            Angle from the circle's center to its first vertex. 
            The default is 2*np.pi/n_points.
        cw: bool, optional
            Clockwise flag for vertex order, the default is True

        Returns
        -------
        Founts

        """
        if start_angle is None and n_points:
            start_angle = 2*np.pi/n_points
        elif not n_points:
            start_angle = 0
        
        # CCW Generation
        if cw:
            angles = start_angle + np.linspace(2*np.pi, 0, n_points, endpoint=False)
        else:
            angles = np.linspace(0, 2*np.pi, n_points, endpoint=False) - start_angle
        
        xs = radius*np.cos(angles)
        ys = radius*np.sin(angles)
            
        kwargs.pop('nodes', None)
        kwargs.pop('expression', None)
        kwargs.pop('domain_min', None)
        kwargs.pop('domain_max', None)
            
        return cls(nodes=np.array([xs, ys]), 
                   expression='polygon',
                   domain_min=-np.abs(radius),
                   domain_max=np.abs(radius),
                   **kwargs)
    
    @classmethod
    def golden(cls,
               n_points: int = 13,
               domain_max: float = 3*np.pi,
               **kwargs) -> typing.Self:
        domain = np.linspace(0, domain_max, n_points)
        
        g  = (1 + 5 ** 0.5) / 2 #golden ratio
        f = g**(domain*g/(2*np.pi)) #factor
        xs = np.cos(domain)*f
        ys = np.sin(domain)*f
        
        kwargs.pop('nodes', None)
        kwargs.pop('expression', None)
        
        return cls(nodes=np.array([xs,ys]),
                   domain_max=domain_max,
                   expression='golden',
                   **kwargs)


def binary_strings_to_list(all_binaries):
    final_output = [[int(b) for b in list(ab)] for ab in all_binaries]
    return(final_output)

def generate_necklace(n = 13):

    #stolen from https://github.com/Ernesti04/necklace_projects/blob/main/necklace_gen_v2.py
    x = 1						# start at 1, avoid all 0s case
    uniques = ["".join(["0"]*n)]
    while (x < 2**n): 			# for each possible number
        s = str(bin(x)[2:].zfill(n)) 	# get binary of number
        cycle = [] 				# get blank list to check rotations
        for i in range(len(s)-1): 	# for each bit in the sequence
            #rot = cycle[i][-1] + cycle[i][:-1] #slightly slower
            rot = s[i:] + s[:i] 		# rotate by i bits
            cycle.append(rot) 		# add each rotation
            if rot < s : 				# if rotation found that is smaller
                break 				# stop searching
        if min(cycle) == s : 		# if the number is already minimum
            uniques.append(s) 		# add to the list
            #print(f'\t{s}') # print results (slow)
        x += 2 					# count odds, halves time
    uniques = binary_strings_to_list(uniques)
    return np.array(uniques)


#functions for drawing lines between points
def non_centre_circle(P: np.ndarray,
                      Q: np.ndarray,
                      b: float = 0.2,
                      samples:int = 150):
    """
    Given a scalar distance b from midpoint of P and Q, define the arc connecting
    P and Q given the calculated center of the circle

    """
    #draws a connecting circle between two points with a centre defined by `b` away from the average. Always chooses the small arc.
    P = np.array(P, dtype=np.float64)
    Q = np.array(Q, dtype=np.float64)
    
    # Calculate new centers
    midpoint = (P + Q) / 2
    vector = P - Q
    mag = np.linalg.norm(vector)
    unit_vector = vector/mag
    
    normal = np.array([unit_vector[1], -unit_vector[0]])
    
    C = midpoint + b*normal
    r = np.linalg.norm(C - P)
    
    theta_P = np.arctan2(P[1] - C[1], P[0] - C[0])
    theta_Q = np.arctan2(Q[1] - C[1], Q[0] - C[0])
    
    # CCW sweep
    ccw_sweep = (theta_Q - theta_P) % (2*np.pi)
    
    # Tangent at P for CCW direction
    v = P - C
    T_ccw = np.array([-v[1], v[0]])  # +90° rotation
    T_ccw /= np.linalg.norm(T_ccw)
    
    # Vector toward origin
    O = -P
    O /= np.linalg.norm(O)
    
    # Dot product tells us if CCW heads toward origin
    dot = np.dot(T_ccw, O)
    
    if dot > 0:
        # CCW heads toward origin → choose CCW arc
        theta = theta_P + np.linspace(0, ccw_sweep, samples)
    else:
        # CW heads toward origin → choose CW arc
        cw_sweep = ccw_sweep - 2*np.pi
        theta = theta_P + np.linspace(0, cw_sweep, samples)

    X = r*np.cos(theta) + C[0]
    Y = r*np.sin(theta) + C[1]
    # plt.figure()
    # plt.plot(X,Y)
    # plt.plot(P, 'o')
    # plt.plot(Q, 'o')
    
    return np.array([X,Y])
    


class Leylines:
    
    __expression_type = None
    PREDEFINED = MappingProxyType({
        'linear' : ('domain', '0*domain'),
        'centre-circle' : ('cos(pi*domain)', 'sin(pi*domain)'),
        # The vector representing a shift in midpoint is 
        # b*np.mean([P,Q],axis=1)/np.hypot(*np.mean([P,Q],axis=1))
        # Given P and Q
        #
        
        
        # TODO
        # Normal towards center
        'non-centre-circle' : non_centre_circle,
        'exponential' : ('domain', 
                         '(exp(10 * domain) - 1) / (exp(10 * domain_max) - 1)'),
        'inverse-exponential': (
            'domain', 
            '-(exp(12 * domain) - 1) / (exp(12 * domain_max) - 1)'),
        })
    PREFEDEFINED_KWARGS = MappingProxyType({
        })
    
    def __init__(self, 
                 founts: Founts = Founts(),
                 expression: str | tuple[str] | typing.Callable = ('domain', '0*domain'),
                 domain_min: float = 0,
                 domain_max: float = 1,
                 resolution: int = 150,
                 **kwargs) -> typing.Self:
        
        n = founts.shape[1]
        self.founts = founts
        # Use numpy array to preserve fancy indexing later
        # (6, 13, 2) is default where 6 orders, 13 founts, and node pairs
        self.line_pairings = np.array([
            [(i, (i + order + 1) % n) for i in range(n)]
            for order in range(int((n-1)/2))])
        self.domain_min = domain_min
        self.domain_max = domain_max
        self.resolution = resolution
        self.kwargs = kwargs
        self.expression = expression
        self.necklace = generate_necklace(n)[1:]
    
    
    @property
    def expression(self) -> str | tuple[str]:
        return self._expression
    
    @expression.setter
    def expression(self, expression: str | tuple[str] | typing.Callable):
        
        
        p_search = lambda s: re.search(r'(\W|\b)P(\W|\b)', s)
        q_search = lambda s: re.search(r'(\W|\b)Q(\W|\b)', s)
        d_search = lambda s: re.search(r'(\W|\b)domain(\W|\b)', s)
        if isinstance(expression, str) and expression in self.PREDEFINED:
            override = self.PREFEDEFINED_KWARGS.get(expression, {}).copy()
            for key in self.kwargs:
                if key in override:
                    override[key] = self.kwargs[key]
            self.kwargs.update(override)
            expression = self.PREDEFINED[expression]
        
        if isinstance(expression, typing.Callable):
            params = inspect.getargs(expression.__code__)
            if 'P' in params.args and 'Q' in params.args:
                self.__expression_type = 'pointwise-callable'
            elif 'domain' in params.args:
                self.__expression_type = 'parametric-callable'
        
        elif isinstance(expression, str):
            if p_search(expression) and q_search(expression):
                self.__expression_type = 'pointwise-str' # singular expression
            elif d_search(expression):
                expression = ("domain", expression)
                self.__expression_type = 'parametric-tuple'
            else:
                raise NotImplementedError(f'\'{expression=:}\' Invalid'
                                          ' form for Leylines!')
                
        elif isinstance(expression, typing.Iterable) and len(expression) == 2:
            if all(p_search(exp) or q_search(exp) for exp in expression):
                self.__expression_type = 'pointwise-tuple'
            elif all(d_search(exp) for exp in expression):
                self.__expression_type = 'parametric-tuple'
            else:
                NotImplementedError(f"'{expression=:}' Invalid form for Leylines")
            expression = tuple(expression)
        
        self._expression = expression
        self.default_curves = self.generate_curves()
        return        
        
    def _eval_parametric(self, samples: int = None) -> np.ndarray:
        if samples is None:
            samples = self.resolution
        domain = np.linspace(self.domain_min, self.domain_max, samples)
        # Parametric function
        env = {'__builtins__': None, 
               **SAFE_MATH, 
               'domain': domain, 
               **self.__dict__,
               **self.kwargs}
        xs = eval(self.expression[0], env)
        ys = eval(self.expression[1], env)
        curve = np.array([xs, ys])
        return curve
    
    def _eval_expression(self, samples: int = None) -> np.ndarray:
        if samples is None:
            samples = self.resolution
        
        domain = np.linspace(self.domain_min, self.domain_max, samples)
        if isinstance(self.expression, str):
            # Simple Function
            env = {'__builtins__': {}, 
                   **SAFE_MATH, 
                   'domain': domain}
            ys = eval(self.expression, env)
            curve = np.array([domain, ys])
        else:
            curve = self._eval_parametric(samples)
            
        return curve
    
    def _normalize_curve(self, curve: np.ndarray) -> np.ndarray:
        """
        Normalizes a curve of n-many samples to the domain [0,1]

        """
        # Shift to start at 0
        curve = curve.copy()
        curve = curve - curve[:, [0]]
        
        # Rotate to basis vector
        dx, dy = curve[0, -1], curve[1, -1]
        angle = -np.arctan2(dy, dx)
        R = np.array([
            [np.cos(angle), -np.sin(angle)],
            [np.sin(angle),  np.cos(angle)]
        ])
        curve = R @ curve
        
        # Normalize to magnitude 1
        curve /= curve[0, -1]

        return curve
    
    def _transform_to_segment(self, 
                              curve: np.ndarray, 
                              P: np.ndarray, 
                              Q: np.ndarray) -> np.ndarray:
        """
        Transforms a normalized curve between two points P and Q

        Parameters
        ----------
        curve : np.ndarray
            Normalized curve - domain[0,1].
        P : np.ndarray
            1x2 array representing ordered pair in x-y coordinate plane.
        Q : np.ndarray
            1x2 array representing ordered pair in x-y coordinate plane.

        Returns
        -------
        curve : np.ndarray
            Curve scaled and rotated to connect P to Q.

        """
        dx = Q[0] - P[0]
        dy = Q[1] - P[1]
        length = np.hypot(dx, dy)
        angle = np.arctan2(dy, dx)

        # Scale
        curve = curve * length

        # Rotate
        R = np.array([
            [np.cos(angle), -np.sin(angle)],
            [np.sin(angle),  np.cos(angle)]
        ])
        curve = R @ curve

        # Translate
        curve = curve + P[:, None]

        return curve

    def generate_curves(self, samples: int = None) -> np.ndarray:
        """
        Generates a Order x Node Count x Ordered Pair x Samples array for every
        possible combination of founts in the self.nodes object evaluated using
        the provided expression.

        Parameters
        ----------
        samples : int, optional
            Number of points to resolve in the curve. The default is self.resolution.

        Returns
        -------
        curves : np.ndarray
            Array of all potential curves given the parameters.

        """
        
        if samples is None:
            samples = self.resolution
        
        # Expressions can be tuple[str] containing "domain"
        #                    tuple[str] containing "P" | "Q"
        #                    str        containing "domain"
        #                    str        containg P & Q
        
        normal = None
        if 'parametric' in self.__expression_type:
            raw_curve = self._eval_expression(samples)
            normal = self._normalize_curve(raw_curve.copy())
        elif 'pointwise' in self.__expression_type:
            pass
        else:
            raise NotImplementedError(f'No idea how you got here, {self.expression=:}, {self.__expression_type=:}')
            
               
        curves = np.zeros((*self.line_pairings.shape,samples), dtype=np.float64)
        
        for order in range(curves.shape[0]):
            for index in range(curves.shape[1]):
                a,b = self.line_pairings[order, index] # indices of founts
                
                P = self.founts[:,a] 
                Q = self.founts[:,b]
                if all(P == Q):
                    continue
                
                if normal is None and 'str' in self.__expression_type: # Pointwise only
                    env = {'__builtins__': __builtins__, 
                            **SAFE_MATH, 
                            'P': P,
                            'Q': Q,
                            'samples': samples,
                            **self.__dict__,
                            **self.kwargs}
                    exec(self.expression, env)
                    curve = np.array([env['X'], env['Y']])
                elif normal is None and 'callable' in self.__expression_type:
                    # Get expression args
                    X,Y = self.expression(P,Q, **{
                        key : value for key, value in self.kwargs.items()
                        if key in inspect.getargs(self.expression.__code__).args
                        and key not in ['P','Q']})
                    curve = np.array([X,Y])
                else:
                    curve = self._transform_to_segment(normal.copy(), P, Q)
                
                # print(f"{P} -> {Q}\n{curve}\n\n")
                curves[order, index] = curve.astype(np.float64)
    
        return curves
    
    def preview(self, order: int = None):
        plt.figure()
        cmap = plt.get_cmap('tab20')
        if order is None:
            for order, order_array in enumerate(self.default_curves):
                for diff, coords in enumerate(order_array):
                    plt.plot(*coords, color=cmap.colors[order])
        else:
            order_array = self.default_curves[order]
            for diff, coords in enumerate(order_array):
                plt.plot(*coords, color=cmap.colors[order])
        
        plt.plot(*self.founts[:,1:], 'bo')
        plt.plot(*self.founts[:,0], 'ro')
        




# -*- coding: utf-8 -*-
"""
Created on Wed Mar 18 20:26:26 2026

@author: trent
"""

#
from pathlib import Path
import warnings
from threading import Thread
import zipfile
import typing
from types import MappingProxyType

import requests
from rapidfuzz import process

from glyphloom.generation import SpellData

#%% 5e.Tools Mirror
try:
    URL = r'https://raw.githubusercontent.com/5etools-mirror-3/5etools-src/main/data/spells'
    SPELLS_BY_SOURCE = {}
    SOURCE_BY_SPELL = {}
    SOURCES = requests.get(f"{URL}/index.json").json()
    def get_spell_map(source:str, file:str) -> None:
        file = f'{URL}/{file}'
        data = requests.get(file).json()
        spell_names = [spell['name'] for spell in data['spell']]
        SPELLS_BY_SOURCE[source] = tuple(spell_names)
        return
    threads = []
    for source, file in SOURCES.items():
        thread = Thread(target=get_spell_map, args=(source,file))
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()
    
    
    SOURCE_BY_SPELL = {spell: 'XPHB' for spell in SPELLS_BY_SOURCE['XPHB']}
    for source, spells in reversed(SPELLS_BY_SOURCE.items()):
        if source == 'PHB':
            continue
        for spell in spells:
            if spell in set(SOURCE_BY_SPELL):
                continue
            SOURCE_BY_SPELL[spell] = source
    del threads, thread, source, file, spell, spells
except Exception as e:
    print(f"{type(e)}: {e}")
    warnings.warn("No access to the 5e.tools Mirror Repo!")
    SOURCES = None
    


#%% Data
FIVE_E = Path(__file__).absolute().parent
SpellData_5e = SpellData.from_yaml(
    (FIVE_E / 'fifth_edition.yaml'))

OFFLINE_LIBRARY = {}
def get_offline_map(file:typing.IO[str]) -> None:
    data = SpellData_5e.yaml_spell(file)
    OFFLINE_LIBRARY[data.name] = data
    return
threads = []
with zipfile.ZipFile(FIVE_E / "library.zip") as zf:
    for file_info in zf.filelist:
        t = Thread(target=get_offline_map, args=(zf.open(file_info),))
        t.start()
        threads.append(t)
    for t in threads:
        t.join()
del zf, t, threads

#%% Monkey Type

AOE_MAP = MappingProxyType({
        "C" : "Cube",
        "E" : "Emanation",
        "H" : "Hemisphere",
        "L" : "Line",
        "MT" : "Multiple Targets",
        "N" : "Cone",
        "Q" : "Square",
        "R" : "Circle",
        "S" : "Sphere",
        "ST" : "Single Target",
        "W" : "Wall",
        "Y" : "Cylinder"
    })
def _get_spell_json_online(spell_name:str, source:str = None) -> SpellData_5e:
    """
    Utilizes the 5etools-mirror-3 repo to load in a spell to the SpellData_5e 
    type

    Parameters
    ----------
    spell_name : str
        Spell name as it appears in the compendium.

    Returns
    -------
    SpellData_5e

    """
    lower, conf, idx = process.extractOne(spell_name.lower(),
                                          map(str.lower, SOURCE_BY_SPELL))
    spell_name = tuple(SOURCE_BY_SPELL)[idx]
    if source is None:
        file = SOURCES[SOURCE_BY_SPELL[spell_name]]
    else:
        file = SOURCES[source]
    url = f'{URL}/{file}'
    data = requests.get(url).json()
    
    spell = max(data['spell'],
                key=lambda sp: sp['name'] == spell_name)
    
    level = spell['level']
    school = max(SpellData_5e.school.options,
                 key=lambda s : s[0] == spell['school'])
    
    damage_type = list(map(str.capitalize, spell.get('damageInflict', [])))
    if len(damage_type) == 1:
        damage_type = damage_type[0]
    elif len(damage_type) > 1:
        damage_type = tuple(sorted(damage_type))
    if not damage_type:
        damage_type = 'None'
    
    aoe_shape = spell.get('areaTags', [])
    aoe_shape = list(map(lambda s : AOE_MAP[s],
                         aoe_shape))
    if len(aoe_shape) == 1:
        aoe_shape = aoe_shape[0]
    elif len(aoe_shape) > 1:
        aoe_shape = tuple(sorted(aoe_shape))
    if not aoe_shape:
        aoe_shape = 'None'
    
    rg = spell['range']
    tp = rg['type']
    if 'distance' in rg:
        if 'amount' in rg['distance']:
            am = rg['distance']['amount']
            unit = rg['distance']['type']
            if am != tp:
                rg = f"{rg['type']} ({am} {unit})"
            else:
                rg = 'point (None)'
        else:
            second = rg['distance']['type'].capitalize()
            first = rg['type']
            rg = f'{first} ({second})'
            # types[tp].add(rg['type'])
    else:
        rg = 'special (Special)'    
            
    duration = spell['duration']
    if len(duration) > 1:
        if duration[0]['type'] == 'instant':
            duration = duration[1]
    else:
        duration = duration[0]
    if duration['type'] == 'timed':
        timing = f'({duration["duration"]["amount"]} {duration["duration"]["type"]})'
        if duration.get('concentration', False):
            duration = f'concentration {timing}'
        else:
            duration = f'timed {timing}'
    elif 'amount' in duration:
        duration = f'timed ({duration["amount"]} {duration["type"]})'
    else:
        duration = f'special ({duration["type"].capitalize()})'
    
    #Ritual
    meta = spell.get('meta', {})
    ritual = False
    if 'ritual' in meta:
        ritual = meta['ritual']
    
    spell_data = {'name' : spell_name,
                  'system': '5e',
                  'level' : level,
                  'school' : school,
                  'damage_type' : damage_type,
                  'area_of_effect' : aoe_shape,
                  'range' : rg,
                  'duration' : duration,
                  'ritual' : ritual}
    for key, value in spell_data.items():
        if value is None:
            spell_data[key] = 'None'
    
    return spell_data
    
    


__SPELL_CACHE = {}

@classmethod
def get_spell(cls, 
              spell_name: str,
              source: str = 'offline') -> SpellData_5e:
    """
    Retrieves a spell from 5eTools using 5.5e PHB, Tasha's, and Xanathar's.
    Spells err on the side of the player's handbook, so if you want the spell
    from a different book, use the source argument with 'xge', 'tce', or the
    last part of the "spells-*.json" data file name from 5e.toosl.data/spells.

    Parameters
    ----------
    spell_name : str
        Name exactly as it appears on 5eTools.
    source: str
        Override command to grab a different book's spells, literal "offline"
        will use the preloaded spells in the library

    Returns
    -------
    SpellData_5e
        Spell Data for the retrieved spell.

    """
    
    if source == 'offline':
        choices = list(OFFLINE_LIBRARY)
        lower, conf, idx = process.extractOne(spell_name.lower(), map(str.lower,choices))
        return tuple(OFFLINE_LIBRARY.values())[idx]
    
    
    if source == 'online' or source is None:
        spell_data_dict = _get_spell_json_online(spell_name)
    elif source:
        spell_data_dict = _get_spell_json_online(spell_name, source)
    
    spell = cls(**spell_data_dict)
    
    return spell
setattr(SpellData_5e, 'get_spell', get_spell)

